from midiPackage.MidiInFile import MidiInFile
import pygame

class baseMidi(object):
    """
    Handles basic music playback.  Subclass it to add midi parsing with a handler.
    """
    
    def __init__(self, midiFile):
        
        self.midiFile = midiFile
        self.paused = False
        pygame.init()
        if pygame.mixer.init():
            print "Mixer didn't initialize"
            import sys
            sys.exit(1)
            
        pygame.mixer.music.load(midiFile)
        
    def play(self):
        pygame.mixer.music.play()
        
    def pause(self):
        if not self.paused:
            pygame.mixer.music.pause()
            self.paused = True
        else:
            pygame.mixer.music.unpause()
            self.paused = False
            
        
    def stop(self, fade=0):
        pygame.mixer.music.fadeout(fade)
        
from midiPackage.MidiOutStream import MidiOutStream

class baseHandler(MidiOutStream):
    """
    Basic functions for the parser.  Subclass it to add parsing capabilities.
    """
    
    def __init__(self):
        MidiOutStream.__init__(self)
        self.length = 0
        self.offset = 0
        self.foundFirstNote = False
        
    def note_on(self, channel=0, note=0x40, velocity=0x40):
        if not self.foundFirstNote:
            self.offset = self.abs_time() * self.realtime
            self.foundFirstNote = True            
        
    def header(self, format=0, nTracks=1, division=91):
        self.division = division
        
    def tempo(self, value):
        self.tempo = value
        self.realtime = self.tempo / (self.division*1000.)
        
    def time_signature(self, nn, dd, cc, bb):
        self.timesig = (nn, 2**dd, cc, bb)
    
    def end_of_track(self):
        end = (self.abs_time() * self.realtime) - self.offset
        if end > self.length:
            self.length = end
