# Very first import b/c it loads a lot of Panda stuff (they're C++ programmers)
import direct.directbase.DirectStart

from time import time
import os
from sets import Set as set
import random
import sys

from direct.showbase.DirectObject import DirectObject
from direct.task.Task import Task
from direct.gui.OnscreenText import OnscreenText
from direct.gui.DirectGui import *
from pandac.PandaModules import TextNode, Vec4

from dancer import dancer
from neat import neatPop, libneat
from picker import picker
from file_browser import Browser
from time import time
from songlist import songList

SONGS = {0:'DrumBea1', 1:'maries', 2:'jazz002', 3:'jazz003', 
                 4:'jazz004', 5:'jazz005', 6:'jazz006', 7:'jazz007', 
         8:'jazz008', 9:'stageclear2', 10:'hotelcalifornia', 
         11:'cometogether'}




#def songs_in_path(path, offset=0):
#    """ returns a dict of songs in path.
#
#    offset is used to make keys start at value other than 0 """
#    songs = [path+song for song in os.walk(path).next()[2]]
#    return dict([(i+offset, song) for i,song in enumerate(songs)])

#SONGS = songs_in_path('songs/legal/', 1)
#SONGS.update(songs_in_path('songs/other/', len(SONGS.keys())))
# try to move song 10 to 0, so it makes sense on keyboard
#SONGS[0] = SONGS.get(10, None)

class baseWorld(DirectObject):
    """
    Sets up everything graphical except the NEAT task to control dancers.
    """
    
    def __init__(self, midiClass, numDancers=1, pos=((0,-5,-13),)):
        
        base.disableAllAudio()
        self.self = self
        f = open("credits.txt", "r")
        self.credits = f.read()
        self.credits = OnscreenText(text = self.credits,
                                    pos=(-1.1,.8), fg=(1,1,1,1),
                                    scale=.05, bg=(0,0,0,.6),
                                    align=TextNode.ALeft)
        self.viewCredits = self._show_credits()
        f.close()
        #taskMgr.doMethodLater(5,self.credits.hide,"hide credits", extraArgs = [])
        self.numDancers = numDancers
        self.midiClass = midiClass
        self.numGens = 0
        self.positions = pos
                         
        self._initPicker()
        self._initEnv()
        self.pickSong(1, withMusic=False)
        self.reload(withMusic=False)
        self.songs = songList(self)
        
        self.forhelp = OnscreenText(text="For help, press [i]",
                                    style=1, fg=(1,1,1,1),
                                    pos=(-0.99,-0.95), scale = .07)
        self.save = OnscreenText(text="Click on the dancer you want to save",
                                 style=1, fg=(1,1,1,1),
                                 pos=(-0.75, -0.95), scale=.08)
        self.save.hide()
        self.load = OnscreenText(text="Click on the dancer to get the saved genome",
                                 style=1, fg=(1,1,1,1),
                                 pos=(-0.5, -0.95), scale=.08)
        self.load.hide()
        

        self.help = self._showHelp()
        self.helpText = self._initKeys()
        self.wantToSave = False
        self.wantToLoad = False
        
        self.mutation_slider = DirectSlider(
            value = libneat.cvar.weight_mut_power, 
            range = (0,3),
            command = self.change_mut_power, 
            pos = (-.95, 0, -.8),
            text = ('Weight Mutation Power = ' + 
                    str(libneat.cvar.weight_mut_power)), 
            text_scale = .2,
            scale = .3, 
            text_pos = (0, .15), 
            textMayChange = 1)
        self.load_from_saved()
        # Must be last call b/c it depends on things other methods set up
        #self.help = self._showHelp()
        self.help.next()
    def _show_credits(self):
        while True:
            yield self.credits.hide()
            yield self.credits.show()
        
    def load_from_saved(self):
        def apply_genome((organism, dancer), genome):
            organism.gnome = libneat.Genome.new_Genome_load(genome)
            organism.update_phenotype()
            dancer.origin = genome
            
        folder = 'genomes/'
        # Get files from the folder
        files = os.walk(folder).next()[2]
        pick_first = [f for f in files if f.startswith('wow')]

        # Divide by 2 so dancers are synced around center
        num = (self.numDancers + 1) // 2

        if len(pick_first) >= num:
            genomes = random.sample(pick_first, num)
        else:
            genomes = pick_first
            the_rest = set(files).difference(set(pick_first))
            how_many_more = min(len(the_rest), num - len(genomes))
            genomes.extend(random.sample(list(the_rest), how_many_more))

        genomes = iter([folder+file for file in genomes])

        orgs = self.neatPop.pop.organisms
        dancers = self.dancers
        pairs = iter([(org, dancer) for org, dancer in zip(orgs, dancers)])

        try:
            try:
                dummy = True
                if len(dancers) % 2 != 0:
                    apply_genome(pairs.next(), genomes.next())
                for gen in genomes:
                    for x in range(2):
                        apply_genome(pairs.next(), gen)
            except StopIteration:
                pass
        finally:
            if not hasattr(self, 'toggle'):
                self.toggle = self.toggle_showing_origin()
                self.toggle.next()
            self.show_origin()

    def toggle_showing_origin(self, force=False):
        self.showing_origin = False
        yield 'ready'
        while 1:
            self.showing_origin = True
            for dancer in self.dancers:
                origin = dancer.origin
                # Remove folders, just get file name
                for slash in ('/', '\\'):
                    if origin.find(slash) != -1:
                        last_slash = ''.join(reversed(origin)).find(slash)
                        origin = origin[-last_slash:]
                t = TextNode(str(dancer))
                t.setText(origin)
                text = dancer.actor.attachNewNode(t)
                text.setScale(8)
                text.setX(text, -.2)
                text.setZ(text, 6.5)
                dancer.text = text
            yield 'hiding filenames in 3 seconds'

            self.showing_origin = False
            for d in self.dancers:
                if hasattr(d, 'text'):
                    d.text.hide()
            yield 'filenames hidden'

    def show_origin(self):
        if self.showing_origin:
            # Do it twice to reload text
            for x in range(2):
                self.self.toggle_showing_origin().next()
        else:
            self.self.toggle_showing_origin().next()
        taskMgr.doMethodLater(3, self.hide_origin, 'hide filenames in 3 seconds')

    def hide_origin(self, *args):
        if self.showing_origin:
            self.self.toggle_showing_origin().next()

    def _initKeys(self):
        """
        Binds keys & mouse clicks to callback functions.
        """
        #for num in range(10):
        #    self.accept(str(num), self.pickSong, [num])

        KEYS = [('space', "Start music", self.midi.play),
                ('enter', "Load random dancers and restart music", self.reload), 
                ('mouse1',
                 "Evolve new dancers by mutating chosen dancer", 
                 self.click),
                ('mouse3',
                 "Evolve new dancers by mating each brain with chosen one",
                 self.rclick),
                # Pause isn't working on Windows
                #('p', "Pause the music", self.midi.pause),
                ('i', "Show/hide help", self.help.next),
                ('s', "Save a dancer", self.saveDancer),
                ('l', "Load a saved dancer", self.loadDancer),
                ('r',     
                 "Reload from saved dancers (like at startup)",
                 self.load_from_saved),
                ('f',     
                 "Show/hide filenames dancers were loaded from", 
                 self.self.toggle_showing_origin().next),
                ('c',
                 "Show credits",
                 self.viewCredits.next),
                ('escape',"Quit", sys.exit)
               ]
        instructions = ['[0-9]: Pick a new song']
        for key, text, callback in KEYS:
            self.accept(key, callback)
            s = ''.join(['[', key, ']: ', text])
            instructions.append(s)
        i = '\n'.join(instructions).replace('mouse1', 'left click')\
                                   .replace('mouse3', 'right click')
        return OnscreenText(i, pos=(-.55, -.25), fg=(1,1,1,1), scale=.05,
                            bg=(0,0,0,.6), align=TextNode.ALeft)

    def _showHelp(self):
        #help = self._initKeys()
        while True:
            yield self.helpText.hide()
            yield self.helpText.show()
            
    def pickSong(self, song=0, name=None, withMusic=True):
        """
        Callback for song selection.  Subclass this to initialize
        any song-specific variables your program needs.
        """
        if name is None:
            song = SONGS[song]
        else:
            song = name
        print song
        self._stopSong()
        self.midi = self.midiClass(song)
        if withMusic:
            self.midi.play()
            
    def reloadDancers(self, _except=None):
        """
        Reloads the dancer models, but skips dancer #_except.
        Subclass this to reposition the models.
        """
        try: self.dancers
        except AttributeError:
            self.dancers = [None] * self.numDancers
        else:#TODO: is this necessary?
            for d,i in zip(self.dancers, range(self.numDancers)):
                if i is not _except:
                    d.actor.delete()
                    
        
        for x in range(len(self.dancers)):
            if x is not _except: #TODO: can this be shortened?
                _dancer = dancer('models/pilot/pilot-model',
                                 {'idle':'models/pilot/pilot-idle'},
                                 'models/pilot/pilotJoints.txt',
                                 rom = 30,
                                 scale = .15,
                                 attach = {'sun':('models/sun', .15, (20,0,420,90,0,0)),
                                           'pointer':('models/toyball', 11, (.2,0,5.1,0,0,0))})
                _dancer.actor.setPos(8 * ((-1)**(x+1)) * ((x+1)/2), 50, -1)
                _dancer.actor.setPos(_dancer.actor, *self.positions[x])
                _dancer.start()
                self.picker.makePickable(_dancer, str(x))
                self.dancers[x] = _dancer

        
    def browserTask(self, browser, dancer, action):
        if not browser.app._quit:
            browser.app.loop()
            if browser.app.file:
                print browser.app.file
                if action == 'save':
                    self.saveDancer(dancer, browser.app.file)
                else:
                    self.loadDancer(dancer, browser.app.file)
                browser.cleanup()
                return Task.done
        else:
            return Task.done

        return Task.cont
    
    def saveDancer(self, dancer=None, file=None):
        if self.wantToLoad:
            return
        elif dancer is None and self.wantToSave: # Pressing 's' again, cancel saving
            print 'foo'
            self.wantToSave = False
            self.forhelp.show()
            self.save.hide()
        elif dancer is None: # Pressing 's' first time
            self.wantToSave = True
            self.forhelp.hide()
            self.save.show()
        elif file is None: # Dancer has been chosen, now pick file
            print 'saving dancer', str(dancer)
            browser = Browser()
            browser.setup()
            taskMgr.add(self.browserTask, 'get save file', 
                        extraArgs=[browser, dancer, 'save'])
        else: # Save the genome to the chosen file
            self.neatPop.pop.organisms[dancer].gnome.print_to_filename(file)
            self.wantToSave = False
            self.save.hide()
            self.forhelp.show()

    def loadDancer(self, dancer=None, file=None):
        if self.wantToSave: 
            return
        elif dancer is None and self.wantToLoad: # Pressing 'l' again, cancel loading
            self.wantToLoad = False
            self.forhelp.show()
            self.load.hide()
        elif dancer is None: # Pressing 'l' first time
            self.wantToLoad = True
            self.forhelp.hide()
            self.load.show()
        elif file is None: # Dancer has been chosen, now pick file
            print 'loading to dancer', str(dancer)
            browser = Browser()
            browser.setup()
            taskMgr.add(self.browserTask, 'get load file',
                        extraArgs=[browser, dancer, 'load'])
        else: # Load the genome from the chosen file
            organism = self.neatPop.pop.organisms[dancer]
            organism.gnome = libneat.Genome.new_Genome_load(file)
            organism.update_phenotype()
            self.dancers[dancer].origin = file
            self.wantToLoad = False
            self.load.hide()
            self.forhelp.show()
            self.show_origin()
            
    def _initEnv(self):
        """ Just loads the disco for now. """
        from disco import disco
        self.disco = disco()
        
    def _initPicker(self):
        """ Sets up the picking callback & task """
        self.picker = picker(self)
        self.hover = []
        taskMgr.add(self._mouseTask, 'mouse watcher')
        
    def _mouseTask(self, task):
        """
        Panda Task that makes the ball appear when the 
        mouse is hovering over a dancer.
        """
        if base.mouseWatcherNode.hasMouse():
            self.picker.getObjectHit(base.mouseWatcherNode.getMouse())
            if self.picker.pickedTag is not None:
                picked = eval(self.picker.pickedTag)
                self.dancers[picked].pointer.show()
                self.hover.append(picked)
            elif len(self.hover):
                for i in self.hover:
                    self.dancers[i].pointer.hide()
                    self.hover.remove(i)
        return Task.cont
        
    def click(self):
        """
        Mouse click callback; population evolves if a dancer is clicked on.
        """
        self.picker.getObjectHit(base.mouseWatcherNode.getMouse())
        if self.picker.pickedTag is not None:
            picked = eval(self.picker.pickedTag)
            try:
                if self.wantToSave:
                    print 'self.wantToSave = True'
                    print picked
                    self.saveDancer(dancer=picked)
                    return
            except AttributeError:
                pass
            try:
                if self.wantToLoad:
                    print 'self.wantToLoad = True'
                    print picked
                    self.loadDancer(dancer=picked)
                    return
            except AttributeError:
                pass
            self.reloadDancers(_except=picked)
            self.dancers[picked].sun.show()
            taskMgr.doMethodLater(1.5, 
                                  self.unpick,
                                  'remove picked notifier', 
                                  extraArgs=[picked])
            self.neatPop.evolve(picked, self.numGens)
            self.numGens += 1
            # Update the origin
            for idx, d in enumerate(self.dancers):
                if idx != picked:
                    d.origin = "mutated"
            self.hide_origin()

    def rclick(self):
        """
        Mouse click callback; population evolves if a dancer is clicked on.
        """
        self.picker.getObjectHit(base.mouseWatcherNode.getMouse())
        if self.picker.pickedTag is not None:
            picked = eval(self.picker.pickedTag)
            self.reloadDancers(_except=picked)
            self.dancers[picked].sun.show()
            taskMgr.doMethodLater(1.5, 
                                  self.unpick, 
                                  'remove picked notifier', 
                                  extraArgs=[picked])
            self.neatPop.mevolve(picked, self.numGens)
            self.numGens += 1
            # Update the origin
            for idx, d in enumerate(self.dancers):
                if idx != picked:
                    d.origin = "crossed"
            else:
                self.hide_origin()
            
    def unpick(self, picked):
        """ Removes the visual cue that a dancer was selected. """
        self.dancers[picked].sun.hide()
        return Task.done
    
    def _stopSong(self, fade=0):
        try:
            self.midi.fade(fade)
            del self.midi
        except AttributeError:
            pass
    
    def change_mut_power(self):
        libneat.cvar.weight_mut_power = self.mutation_slider['value']
        self.mutation_slider['text'] = ('Weight Mutation Power = %.2f' % 
                                        self.mutation_slider['value'])
        
    def reload(self, withMusic=True):
        """ Resets the dancers, loads a new random network population. """
        self.reloadDancers()
        self.neatPop = neatPop(4, 
                               len(self.dancers[0].jointList), 
                               len(self.dancers))
        if withMusic:
            self.midi.play()


if __name__ == '__main__':
    from baseMidi import baseMidi
    w = baseWorld(baseMidi)
    run()
