
# set gui key to None if you want to call toggleConsole from outside this class
GUI_KEY = "f1"

# change size of text and number of characters on one line
# scale of frame ( must be small (0.0x)
SCALE = 0.05

# V_SPACE + V_SIZE should not exceed 2.0, else parts of the interface will not be visible
# space above the frame ( must be below 2.0, best between 0.0 and 1.0 )
V_SPACE = 0.1
# vertical size of the frame ( must be at max 2.0, best between 0.5 and 2.0 )
V_SIZE = 1.8

# when outputting text to the console, carriage return after LINELENGTH characters
LINELENGTH = 80

# do not change unless you change the size of the frames
NUMLINES = int(V_SIZE/SCALE - 3)

# to define a special font, if loading fails the default font is used (without warning)
FONT = 'vera.ttf'

# output debug informations
DEBUG = 0

import sys, re, string

from code import InteractiveConsole
from direct.gui.DirectGui import DirectFrame, DirectEntry, DirectLabel, DGG
from direct.showbase import DirectObject
from pandac.PandaModules import TextNode
from direct.gui.OnscreenText import OnscreenText
from direct.task import Task

if sys.platform == 'win32':
  import msvcrt
elif sys.platform == 'linux2':
  import termios, fcntl, sys, os

# do not change unless you know what you do... there are constants used as parameters
GUI = 1
CONSOLE = 2

class pandaInteractiveConsole( InteractiveConsole, DirectObject.DirectObject ):
  def __init__( self, locals, interfaces=GUI|CONSOLE ):
    if DEBUG:
      print "I: pandaInteractiveConsole.__init__"
   
    InteractiveConsole.__init__( self, locals )
   
    self.guiEnabled = (interfaces & GUI)
    if self.guiEnabled:
      if DEBUG:
        print "I:  - GUI enabled"
     
      if (V_SPACE + V_SIZE) > 2.0:
        print "W:  V_SPACE + V_SIZE > 2.0 is not a wise idea, parts of the interactiveConsole will not be visible"
     
      top = 1/SCALE - (V_SPACE/SCALE)
      bottom = 1/SCALE - (V_SPACE/SCALE) - (V_SIZE/SCALE)
     
      # panda3d interface
      self.consoleFrame = DirectFrame( relief = DGG.GROOVE
        , frameColor = (200, 200, 200, 0.5), scale=SCALE
        , frameSize = (-1/SCALE, 1/SCALE, top, bottom))
     
      # try to load the defined font     
#      fixedWidth = loader.loadFont(FONT)
#      # if font is not valid use default font
#      if not fixedWidth.isValid():
#        if FONT is not None:
#          print "pandaInteractiveConsole.py :: could not load the defined font %s" % str(FONT)
#       fixedWidth = DGG.getDefaultFont()
        
      fixedWidth = DGG.getDefaultFont() 
     
      # text entry line
      self.consoleEntry = DirectEntry( self.consoleFrame, text = "", command=self.setText, width = 1/SCALE*2-2
        , pos =(-1/SCALE+1,0,bottom+1), initialText="", numLines = 1, focus=1, entryFont=fixedWidth)
      # output lines
      self.consoleOutputList = list()
      for i in xrange( NUMLINES ):
        label = OnscreenText( parent = self.consoleFrame, text = "", pos = (-1/SCALE+1, bottom+2+NUMLINES-i)
          , align=TextNode.ALeft, mayChange=1, scale=1.0)
        label.setFont( fixedWidth )
        self.consoleOutputList.append( label )
        self.toggleConsole()
     
      if GUI_KEY is not None:
        self.accept( GUI_KEY, self.toggleConsole )
   
    # redirect input and output
    self.stdout = sys.stdout
    sys.stdout = self
   
    self.consoleEnabled = (interfaces & CONSOLE)
    if self.consoleEnabled:
      if DEBUG:
        print "I:  - Console enabled"
     
      # task to read the command line input
      taskMgr.add(self.read, 'pandaInteractiveConsoleReadtask')
     
      if DEBUG:
        self.write( "D: myInteractiveConsole.__init__\n" )
     
      # buffer of current text
      self.linebuffer = ''
      self.linebufferPos = 0
      if sys.platform == 'linux2':
        # set read non blocking
        fd = sys.stdin.fileno()
        self.oldterm = termios.tcgetattr(fd)
        newattr = termios.tcgetattr(fd)
        newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
        termios.tcsetattr(fd, termios.TCSANOW, newattr)
        self.oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, self.oldflags | os.O_NONBLOCK)
   
 
  # toggle the gui console
  def toggleConsole( self ):
    self.consoleFrame.toggleVis()
    self.consoleEntry['focus'] != self.consoleFrame.isHidden
 
  # set the text of the panda3d console
  def setText(self, textEntered):
    self.write( ">%s" % textEntered )
    output = self.push( textEntered )
    # clear line
    self.consoleEntry.enterText('')
    self.consoleEntry['focus'] = 1
 
  # write a string to the panda3d console
  def writeP3dConsole( self, printString ):
    # remove carriage returns (causes disort in output)
    printString = printString.strip()
    # remove not printable characters (which can be input by console input)
    printString = re.sub( r"[^%s]" % string.printable, "", printString )
    # only write lines which contain something
    if printString.strip() != '':
      while len(printString) > LINELENGTH:
        writeString = printString[0:LINELENGTH]
        printString = printString[LINELENGTH:]
        # output on panda3d interface
        for i in xrange( NUMLINES-1 ):
          self.consoleOutputList[i].setText( self.consoleOutputList[i+1].getText() )
        self.consoleOutputList[NUMLINES-1].setText( writeString )
      # output on panda3d interface
      for i in xrange( NUMLINES-1 ):
        self.consoleOutputList[i].setText( self.consoleOutputList[i+1].getText() )
      self.consoleOutputList[NUMLINES-1].setText( printString )
 
  # write a string to the console (print does not work within this class)
  def write( self, string ):
    try:
      if self.consoleEnabled:
        # output on console
        self.stdout.write( string )
      # output on panda3d interface
      if self.guiEnabled:
        self.writeP3dConsole( string )
    except IOError:
      # this can happen on the linux server, dont know why
      pass
    except UnicodeEncodeError:
      # dont know how to handle this currently (maybe needs manual conversion)
      pass
 
  # restore command line settings
  def __del__( self ):
    if self.consoleEnabled:
      # reset settings of command console
      if sys.platform == 'linux2':
        self.write( "setting old terminal settings\n" )
        termios.tcsetattr(fd, termios.TCSAFLUSH, self.oldterm)
        fcntl.fcntl(fd, fcntl.F_SETFL, self.oldflags)
 
  # clear the line of the command console, (reqired for backspaces)
  def refreshLine( self ):
    # clear complete line
    for i in xrange(len(self.linebuffer)):
      self.write("\b")
    # write it again
    self.write( self.linebuffer+" " )
    self.write("\b")
 
  # read a single keypress from command line, on special key multiple reads are required
  def readKey( self ):
    # read key from keyboard
    char = None
    if sys.platform == 'win32':
      if msvcrt.kbhit():
        char = msvcrt.getch()
    elif sys.platform == 'linux2':
      try:
        char = sys.stdin.read(1)
      except IOError: pass
    return char
 
  def read( self, task ):
    # get keypress
    char = self.readKey()
    while char is not None:
      # clear last character on backspace
      if ord(char) == 8 or ord(char) == 127:
        self.write(char)
        self.linebuffer = self.linebuffer[0:self.linebufferPos-1] + self.linebuffer[self.linebufferPos:len(self.linebuffer)]
        self.linebufferPos = max(0, self.linebufferPos - 1)
        self.refreshLine()
      # execute command on carriage return
      elif ord(char) == 13 or ord(char) == 10: # 13 on windows / 10 on linux ???
        self.write( "\n> exec '%s'" % self.linebuffer )
        print
        output = self.push( self.linebuffer )
        self.linebuffer = ''
        self.linebufferPos = 0
      # move cursor to edit commandline
      elif ord(char) == 224 or ord(char) == 1: # ord(char) == 27
        if sys.platform == 'linux2':
          cmdKey = self.readKey()
        cmdKey = self.readKey()
        if ord(cmdKey) == 75 or ord(cmdKey) == 68:
          self.linebufferPos = max( 0, min( len(self.linebuffer), self.linebufferPos - 1 ) )
        elif ord(cmdKey) == 77 or ord(cmdKey) == 67:
          self.linebufferPos = max( 0, min( len(self.linebuffer), self.linebufferPos + 1 ) )
      # all other keypresses are added to the command line
      else:
        self.linebuffer = self.linebuffer[0:self.linebufferPos] + char + self.linebuffer[self.linebufferPos:len(self.linebuffer)]
        if self.linebufferPos+1 != len(self.linebuffer):
          diffLen = len(self.linebuffer) - self.linebufferPos+1
          self.write(char)
          self.refreshLine()
        else:
          self.write(char)
        self.linebufferPos += 1
      # get keypress
      char = self.readKey()
    return Task.cont 
