from direct.actor.Actor import Actor
from pandac.PandaModules import NodePath

class dancer(object):
    """
    Holds an actor model and his joint list.
    """
    
    def __init__(self, model, anim, jointsFile, rom=0, scale=1, attach={}):     
        """ Creates a new dancer.
        
        @model: the panda egg/bam file of the model
        @anim: (dict) an animation file with an arbitrary key
        @jointsFile: the file used to take control of the joints
        @rom: the minimum range of motion a joint must expose to be controlled by the NN
        @scale: visual scale of the model
        @attach: any objects that should be attached to the dancer
        """

        self.actor = Actor(model, anim)
        self.actor.reparentTo(render)
        self.actor.setScale(scale)
        self.anim = anim
        self.jointsFile = jointsFile
        self.rom = rom
        self.moved = False
        self.origin = "random"
        for k in attach:
            self.attach(k, *attach[k])

    def start(self):
        self.startPos = self.actor.getPos()
        self.jointList = self.readJointsandNarrow(self.actor, self.jointsFile, self.rom)
        self.actor.actorInterval(self.anim.keys()[0], playRate=1).loop()
        del self.jointsFile
        del self.rom
        del self.anim

    
    def attach(self, name, path, scale, posHpr):
        obj = loader.loadModel(path)
        obj.setScale(scale)
        obj.setPosHpr(obj, *posHpr)
        obj.hide()
        obj.reparentTo(self.actor)
        self.__dict__[name] = obj
        
    def readJointsandNarrow(self, actor, file, rom=0):
        """
        Takes control of actor's joints as directed in file, narrows joints if rom > 0.
        
        returns: list of (joint,limits,axis) tuples
        """
        import string
        infile = open(file, 'r')
        jointList = []
        axes = ['H','P','R']
        i = 0
        for line in infile:
            if line[0] != '#':
                name, joint, origin, limits, offset = map(string.strip, line.split(':'))
                
                if i is 4:
                    self.rightFoot = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 5:
                    self.leftFoot = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 2:
                    self.rightKnee = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 3:
                    self.leftKnee = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 0:
                    self.rightHip = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 1:
                    self.leftHip = actor.exposeJoint(None, 'modelRoot', joint)
                elif i is 14:
                    self.pelvis = actor.exposeJoint(None, 'modelRoot', joint)
                joint = actor.controlJoint(None, 'modelRoot', joint)
                i+= 1
                if line[:3] is not 'no-':
                    h,p,r = eval(origin)
                    joint.setHpr(h,p,r)
                    # skipping the offset & name for now
                    limits = eval(limits)
                    for lim,axis in zip(limits, axes):
                        if lim[1] - lim[0] >= rom:
                            jointList.append((joint, lim, axis))
        return jointList
        
    @staticmethod
    def restrain(x, limits):
        return min(max(x, limits[0]), limits[1])

    def lowestPoint(self):
        return min(self.rightFoot.getPos(render)[2], self.leftFoot.getPos(render)[2])
    
    def setHeight(self):
        self.height = self.startPos()[2]-self.lowestPoint()
    
    jointFuncs = {'H':(NodePath.getH, NodePath.setH),
                  'P':(NodePath.getP, NodePath.setP),
                  'R':(NodePath.getR, NodePath.setR)}    
    def move(self, joint, delta):
        """ Move joint by delta, according to joint's limits & axis. """
        get, set = dancer.jointFuncs[joint[2]]
        set(joint[0], self.restrain(get(joint[0])+delta, joint[1]))
        self.fixHeight()

    def fixHeight(self):
        self.actor.setPos(self.startPos[0], self.startPos[1], self.startPos[2]-(self.height-(self.actor.getPos()[2]-self.lowestPoint())))
