import unittest
from itertools import groupby

from midiPackage.MidiInFile import MidiInFile

from baseMidi import baseMidi, baseHandler

class FrequencyFilter(baseHandler):

    def __init__(self):
        baseHandler.__init__(self)
        # Empty dict for each midi channel
        self.notes = dict([(x,{}) for x in range(16)])

    def note_on(self, channel=0, note=0x40, velocity=0x40):
        baseHandler.note_on(self, channel, note, velocity)

        channel = self.notes[channel]
        channel[note] = channel.get(note, 0) + 1

    def heuristic(self):
        """ Filter notes by frequency """
        def tuplify():
            """ convert dict {channel : {note : count}} to tuple (channel, note, count) """
            lst = []
            for k in self.notes:
                for kk in self.notes[k]:
                    lst.append((k, kk, self.notes[k][kk]))
            return lst
        notes = tuplify()
        notes.sort(lambda a,b: cmp(b[2], a[2])) # sort by note count, descending

        fourths = int(len(notes) / 4)
        notes = [(chan, note) for chan, note, cnt in notes[2*fourths : 3*fourths]]

        # Sort by channel so groupby gets all notes of channel n together.
        notes.sort(lambda a,b: cmp(a[0], b[0]))

        note_dict = {}
        # Create dict: {channel : note} by grouping on channel
        for chan, iter_notes in groupby(notes, lambda x: x[0]):
            note_dict[chan] = [x[1] for x in iter_notes]

        return note_dict




# The availiable filters
def Types(dir):
    """ finds the available filters in this module """
    t = []
    for s in dir:
        if s.find('Filter') != -1:
            t.append(s[:s.find('Filter')].lower())
    return t
types = Types(dir())

def get_by(filter, midi_file):
    handler = eval(filter.title() + 'Filter')()
    midi_in = MidiInFile(handler, midi_file)
    midi_in.read()
    return handler.heuristic()

class FilterTestCase(unittest.TestCase):

    def test_heuristic(self):
        song = 'songs/legal/communication.mid'
        channels = range(16)
        notes = range(127)
        for filter in types:
            inputs = get_by(filter, song)
            for k in inputs.keys():
                self.assert_(k in channels)
                for note in inputs[k]:
                    self.assert_(note in notes)

def suite():
    return unittest.MakeSuite(FilterTestCase, 'test')

if __name__ == '__main__':
    unittest.main()
