#!/usr/bin/env python
from decimal import Decimal
import sys

from direct.task.Task import Task
from pandac.PandaModules import *
from direct.gui.DirectGui import *
from pygame.mixer import music
from midiPackage.MidiInFile import MidiInFile
from songlist import songList
from baseWorld import baseWorld
from baseMidi import baseMidi, baseHandler
import handlers

class World(baseWorld):
    
    def __init__(self):
        positions = ((0,-5,-13),(0,30,1),(0,30,1),(-15,-10,-13),(15,-10,-13))                    
        baseWorld.__init__(self, Midi, numDancers = 5, pos=positions)
        taskMgr.add(self.neatTask, 'my new NEAT task')
        self.rightFoot = self.dancers[0].actor.exposeJoint(None, 'modelRoot', 'Bone_r_foot_heel')
        self.leftFoot = self.dancers[0].actor.exposeJoint(None, 'modelRoot', 'Bone_l_foot_heel')
        self.moved = False
        
        buttons = [("Save", self.saveDancer),
                   ("Load", self.loadDancer), 
                   ("Restart", self.reload)]
        for idx,(text, cmd) in enumerate(buttons):
            self.button(text, cmd, idx - len(buttons)//2)
        self.accept("p", self.pause)

    def button(self, text, cmd, x):
        return DirectButton(text=text, 
                            command=cmd,
                            scale=.07,
                            pos=(x*.2 + .5,0,-.95), 
                            textMayChange=True)
    def pause(self):
        self.midi.pause()
        
    def neatTask(self, task):
        if music.get_busy() and not self.midi.paused:
            pos = music.get_pos()
            measure = (pos % self.midi.measure) / self.midi.measure
            beat = (pos % self.midi.beat) / self.midi.beat
            
            #if pos > self.curNote[3]:
            #    try:
            #        self.curNote = self.instrumental.next()
            #    except StopIteration:
            #        self.curNote = [0,0,Decimal('infinity'),0,0]
            #if self.curNote[2] < pos < self.curNote[3]:
            #    instrument = 1 - (pos-self.curNote[2]) / self.curNote[4]
            #else:
            #    instrument = 0

            if pos > self.drums[3]:
                try:
                    self.drums = self.percussion.next()
                except StopIteration:
                    self.drums = [0,0,Decimal('infinity'),0,0]
            if self.drums[2] < pos < self.drums[3]:
                drums = 1 - (pos-self.drums[2]) / self.drums[4]
            else:
                drums = 0

            try: self.heights
            except AttributeError:
                self.heights = [0.0]*self.numDancers
                for x in range(len(self.dancers)):
                    self.heights[x] = self.dancers[x].startPos[2]-self.dancers[x].lowestPoint()
                    self.dancers[x].height = self.heights[x]
                    print self.dancers[x].lowestPoint(), self.heights[x]
            else:
                for x in range(len(self.dancers)):
                    try: self.dancers[x].height
                    except AttributeError:
                        self.dancers[x].height = self.heights[x]
            outputs = self.neatPop.gen((measure, beat, drums, 1.0))
            x = 0
            for tup, dancer in zip(outputs, self.dancers):
                for delta, joint in zip(tup, dancer.jointList):
                    dancer.move(joint, 2*delta)
        elif not self.midi.paused:
            #self.instrumental = iter(self.midi.instruments)
            self.percussion = iter(self.midi.percussion)

        return Task.cont
    
    def pickSong(self, song=0, name="songs/communication.mid", withMusic=True):
        baseWorld.pickSong(self, song, name, False)
        #self.instrumental = iter(self.midi.instruments)
        self.percussion = iter(self.midi.percussion)
        #try:
        #    self.curNote = self.instrumental.next()
        #except StopIteration: pass
        try:
            self.drums = self.percussion.next()
        except StopIteration: pass
        if withMusic:
            self.midi.play()
        
    
class Midi(baseMidi):
    
    def __init__(self, midiFile):
        baseMidi.__init__(self, midiFile)

        # Figure out what notes/channels to watch, and pass that to transposer
        if len(sys.argv) > 1 and sys.argv[1] in handlers.types:
            inputs = handlers.get_by(sys.argv[1], midiFile)
        else:
            inputs = handlers.get_by('frequency', midiFile)
        inputs.pop(9, None) # Get rid of percussion notes
        #self.instruments = Transposer(inputs)
        #midiIn = MidiInFile(self.instruments, midiFile)
        #midiIn.read()
        
        #self.measure = self.instruments.division * self.instruments.timesig[0] * \
         #              (4 / self.instruments.timesig[1]) * self.instruments.realtime
        #self.beat = self.measure / self.instruments.timesig[1]
        #self.instruments = sorted(self.instruments.notes, 
        #                          lambda a,b: cmp(a[2], b[2]))

        self.percussion = Transposer({9:[35,37,38,44,49,69]})
        midiIn = MidiInFile(self.percussion, midiFile)
        midiIn.read()
        self.measure = self.percussion.division * self.percussion.timesig[0] * \
                    (4 / self.percussion.timesig[1]) * self.percussion.realtime
        self.beat = self.measure / self.percussion.timesig[1]
        self.percussion = sorted(self.percussion.notes,
                                 lambda a,b: cmp(a[2], b[2]))
        
           
class Transposer(baseHandler):
    
    def __init__(self, inputs=None):
        baseHandler.__init__(self)
        if inputs is None:
            inputs = {9 : [35,37,38,44,49,69]}
        self.inputs = inputs
        print self.inputs
        self.notes = []
        
    def note_on(self, channel=0, note=0x40, velocity=0x40):
        baseHandler.note_on(self, channel, note, velocity)
        if self.inputs.has_key(channel) and note in self.inputs[channel]:
            start = self.abs_time() * self.realtime - self.offset
            self.notes.append([channel, note, start, 0, 0])
            
    def note_off(self, channel=0, note=0x40, velocity=0x40):
        if self.inputs.has_key(channel) and note in self.inputs[channel]:
            for i in reversed(self.notes):
                if i[:2] == [channel, note]:
                    i[3] = self.abs_time() * self.realtime - self.offset
                    i[4] = (i[3]-i[2])*4
                    return

if __name__ == '__main__':
    w = World()
    from console import pandaInteractiveConsole as pandaConsole, GUI, CONSOLE
    console = pandaConsole(locals(), GUI|CONSOLE)
    run()
